from .lowlevel import __call
from .fasta import FastaParser
from Bio import Phylo
import os


def muscle(fasta, output, maxiters=16, outformat='fasta', log=None):
    """
    Wrapper function for MSUCLE sequence alignment

    Args:
        fasta: str, path-like
            The input fasta file

        output: str, path-like
            The output file

        maxiters: int
            Maximum number of iterations (default 16)

        outformat: str
            'fasta' (default)
            'html'
            'msf' (GCF MSF)
            'clw' (ClustalW)

        log: str, path-like
            The log file, default None -> <output>.log
    """
    outformat = {'fasta': '', 'html': '-html', 'msf': '-msf', 'clw': '-clw'}[outformat]

    if log is None: log = output + '.log'
    __call(f'muscle -in {fasta} -out {output} -maxiters {maxiters} -log {log} {outformat}')


def simple_phylogeny(fasta, output, clustering='NJ', log=None):
    """
    Make a tree file (newick/phylip format) from the aligned <fasta>
        using the command "clustalw -tree"

    Args:
        fasta: str, path-like
            The input fasta file, sequences must be aligned

        output: str, path-like
            The output tree file

        clustering: str
            clustering method, 'NJ' or 'UPGMA'

        log: str, path-like
            The log file, default None -> <output>.log
    """
    with FastaParser(fasta) as parser:
        head, seq = parser.next()
        if set(seq.upper()) == set('ACGT-'): type_ = 'DNA'
        else: type_ = 'PROTEIN'

    if log is None: log = output + '.log'
    __call(f'clustalw -tree -infile={fasta} -type={type_} -CLUSTERING={clustering} > {log}')

    # Rename the .ph file automatically generated by clustalw
    ph_file = fasta[:fasta.rfind('.')] + '.ph'
    os.rename(ph_file, output)


def draw_tree(file, ax):
    """
    Wrapper function for Bio.Phylo.read() and Bio.Phylo.draw()
        to draw a phylogenetic tree from the input tree <file>

    Args:
        file: str, path-like
            The input tree file in newick/phylip format

        ax: matplotlib Axes object
            The ax object in which the tree will be drawn
    """
    tree = Phylo.read(file, 'newick')
    Phylo.draw(tree, axes=ax)

